/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGraphicsScene>
#include <QGraphicsProxyWidget>
#include <QTransform>
#include <QTimeLine>
#include <QEvent>
#include <QWheelEvent>
#include <QMouseEvent>
#include <QPaintEvent>
#include <QEvent>
#include <assert.h>


#include "gridwidget.h"
#include "vector2.h"

TGridWidgetScene::TGridWidgetScene(
	const QObject 						*parent)
:QGraphicsScene(const_cast<QObject*>(parent))
{
	d_animation_mode = false;
}

TGridWidgetScene::~TGridWidgetScene(void)
{

}

void TGridWidgetScene::drawForeground(
	QPainter 							*painter,
	const QRectF 						&exposed)
{
	if(d_animation_mode == true && d_Scene_Image.isNull() == false)
	{
		painter->drawPixmap(exposed, d_Scene_Image, exposed);
	}
}

void TGridWidgetScene::drawItems(
	QPainter 							*painter,
	int 								numItems,
	QGraphicsItem 						*items[],
	const QStyleOptionGraphicsItem 		options[],
	QWidget 							*widget)
{
	if(!d_animation_mode)
	{
		QGraphicsScene::drawItems(painter,numItems,items,options,widget);
	}
}

TGridWidget::TGridWidget(
	const QWidget 						*Parent)
:QGraphicsView(const_cast<QWidget*>(Parent)),
 d_AnimationRunning(false),
 d_MaxColumns(1),
 d_InitialResizingDone(false),
 d_ActivateWidgetQueued(false),
 d_CurrentActiveWidgetIndex(-1),
 d_delayed_remove_widget_active(false)
{

	this->setOptimizationFlags(QGraphicsView::DontAdjustForAntialiasing | QGraphicsView::IndirectPainting);

    this->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    this->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

	d_GraphicsScene = new TGridWidgetScene(this);

	this->setScene(d_GraphicsScene);
	this->setTransformationAnchor(QGraphicsView::NoAnchor);

	d_AnimationTimeLine = new QTimeLine(500,this);
	d_AnimationTimeLine->setCurveShape(QTimeLine::EaseInOutCurve);
	d_AnimationTimeLine->setUpdateInterval(35);

	this->setBackgroundBrush(this->palette().window());

	connect(d_AnimationTimeLine,SIGNAL(valueChanged(qreal)),this,SLOT(AnimationValueChanged(qreal)),Qt::UniqueConnection);
	connect(d_AnimationTimeLine,SIGNAL(finished(void)),this,SLOT(AnimationFinished(void)),Qt::UniqueConnection);
}

TGridWidget::~TGridWidget(void)
{

}

QWidget* TGridWidget::ActiveWidget(void) const
{
	if(d_Widgets.size())
	{
		if(static_cast<unsigned int>(d_CurrentActiveWidgetIndex) < d_Widgets.size())
		{
			return d_Widgets[d_CurrentActiveWidgetIndex]->widget();
		}
	}

	return 0;
}

void TGridWidget::AddWidget(
	QWidget 							*Widget,
	Qt::WindowFlags 					Flags)
{
	QGraphicsProxyWidget				*ProxyWidget;
	QSizePolicy 						SizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);

	SizePolicy.setHorizontalStretch(0);
	SizePolicy.setVerticalStretch(0);

	assert(Widget);

	ProxyWidget = d_GraphicsScene->addWidget(Widget,Flags);
	SizePolicy.setHeightForWidth(ProxyWidget->sizePolicy().hasHeightForWidth());
	ProxyWidget->setSizePolicy(SizePolicy);
	d_Widgets.push_back(ProxyWidget);

	if(d_Widgets.size() == 1)
	{
		d_CurrentActiveWidgetIndex = 0;					// First widget is active by default
		d_WidgetSize = Widget->size();					// Default size is from first widget
		d_current_viewport.setRect(0,0,d_WidgetSize.width(),d_WidgetSize.height());	// Default viewport around first widget
	}
}

void TGridWidget::RemoveWidget(
	const QWidget 						*widget)
{
	d_delayed_remove_widgets.push_back(widget);
	d_delayed_remove_widget_active = true;
}

void TGridWidget::Resize_Widgets(void)
{
	std::vector<QGraphicsProxyWidget*>::iterator IterProxy;
	TVector2							Offset;
	int									ColumnCntr;
	QPointF								Point;
	QPointF								pnt1,pnt2;
	int									row;
	int									column;
	QRectF								scene_viewport;

	d_WidgetSize = this->size();

	if(d_Widgets.size() < 3)
	{
		d_MaxColumns = 2;
	}
	else
	{
		d_MaxColumns = 0.5 + sqrt(static_cast<double>(d_Widgets.size()));
	}

	ColumnCntr = 0;
	Offset.Set(0,0);

	for(IterProxy = d_Widgets.begin();IterProxy != d_Widgets.end();IterProxy++)
	{
		(*IterProxy)->resize(d_WidgetSize);
		(*IterProxy)->setPos(Offset.x,Offset.y);

		ColumnCntr++;

		if(ColumnCntr < d_MaxColumns)
		{
			Offset.x += d_WidgetSize.width();
		}
		else
		{
			Offset.x = 0;
			Offset.y += d_WidgetSize.height();
			ColumnCntr = 0;
		}
	}

	// Find scene viewport
	if(d_Widgets.size())
	{
		row = (d_Widgets.size() - 1) / d_MaxColumns;

		pnt1.setX(0);
		pnt1.setY(0);

		pnt2.setX(d_WidgetSize.width() * d_MaxColumns);
		pnt2.setY(d_WidgetSize.height() * (row + 1));

		scene_viewport.setTopLeft(pnt1);
		scene_viewport.setBottomRight(pnt2);

		d_GraphicsScene->setSceneRect(scene_viewport);
	}


	if(static_cast<unsigned int>(d_CurrentActiveWidgetIndex) < d_Widgets.size())	// adjust current viewport
	{
		row = d_CurrentActiveWidgetIndex / d_MaxColumns;
		column = d_CurrentActiveWidgetIndex - row * d_MaxColumns;

		pnt1.setX(d_WidgetSize.width() * column);
		pnt1.setY(d_WidgetSize.height() * row);

		pnt2.setX(d_WidgetSize.width() * (column + 1));
		pnt2.setY(d_WidgetSize.height() * (row + 1));

		d_current_viewport.setTopLeft(pnt1);
		d_current_viewport.setBottomRight(pnt2);

		this->fitInView(d_current_viewport,Qt::KeepAspectRatio);

	}
	else if(d_Widgets.size())
	{
		d_current_viewport = scene_viewport;

		this->fitInView(d_current_viewport,Qt::KeepAspectRatio);
	}

	if(!d_InitialResizingDone)
	{
		d_InitialResizingDone = true;

		if(d_ActivateWidgetQueued)
		{
			this->ActivateWidget(d_InitialActivateWidget);
		}
	}
}

void TGridWidget::ActivateWidget(
	const int 							Index)
{
	int									Row;
	int									Column;
	QPointF								pnt1,pnt2;

	if(d_AnimationRunning)
	{
		return;
	}

	if(!d_Widgets.size())
	{
		return;
	}

	if(!d_InitialResizingDone)
	{
		d_ActivateWidgetQueued = true;
		d_InitialActivateWidget = Index;

		return;
	}

	this->resetTransform();

	if(this->isVisible())
	{
		this->Create_Background_Image();
	}

	if(static_cast<unsigned int>(Index) < d_Widgets.size())
	{
		// Show selected widget
		Row = Index / d_MaxColumns;
		Column = Index - Row * d_MaxColumns;

		pnt1.setX(d_WidgetSize.width() * Column);
		pnt1.setY(d_WidgetSize.height() * Row);

		pnt2.setX(d_WidgetSize.width() * (Column + 1));
		pnt2.setY(d_WidgetSize.height() * (Row + 1));

		d_target_viewport.setTopLeft(pnt1);
		d_target_viewport.setBottomRight(pnt2);

		d_CurrentActiveWidgetIndex = Index;

		d_GraphicsScene->Set_Animation_Mode(true);
		d_AnimationTimeLine->start();
		d_AnimationRunning = true;

		return;
	}

	d_CurrentActiveWidgetIndex = -1;

	Row = (d_Widgets.size() - 1) / d_MaxColumns;

	pnt1.setX(0);
	pnt1.setY(0);

	pnt2.setX(d_WidgetSize.width() * d_MaxColumns);
	pnt2.setY(d_WidgetSize.height() * (Row + 1));

	d_target_viewport.setTopLeft(pnt1);
	d_target_viewport.setBottomRight(pnt2);

	d_GraphicsScene->Set_Animation_Mode(true);
	d_AnimationTimeLine->start();
	d_AnimationRunning = true;

}

void TGridWidget::ActivateWidget(
	const QWidget 						*Widget)
{
	std::vector<QGraphicsProxyWidget*>::const_iterator IterProxy;
	int									Offset;

	assert(Widget);

	Offset = 0;
	for(IterProxy = d_Widgets.begin();IterProxy != d_Widgets.end();IterProxy++)
	{
		if(Widget == (*IterProxy)->widget())
		{
			this->ActivateWidget(Offset);
			return;
		}

		Offset++;
	}

	this->ActivateWidget(-1);
}

void TGridWidget::AnimationValueChanged(
	qreal								Value)
{
	std::vector<QGraphicsProxyWidget*>::iterator IterProxy;
	QRectF								Viewport;
	QPointF								start_pnt1,start_pnt2;
	QPointF								end_pnt1,end_pnt2;
	QPointF								inter_pnt1,inter_pnt2;

	this->setUpdatesEnabled(false);

	start_pnt1 = d_current_viewport.topLeft();
	start_pnt2 = d_current_viewport.bottomRight();

	end_pnt1 = d_target_viewport.topLeft();
	end_pnt2 = d_target_viewport.bottomRight();

	inter_pnt1.setX(start_pnt1.x() + (end_pnt1.x() - start_pnt1.x()) * Value);
	inter_pnt1.setY(start_pnt1.y() + (end_pnt1.y() - start_pnt1.y()) * Value);

	inter_pnt2.setX(start_pnt2.x() + (end_pnt2.x() - start_pnt2.x()) * Value);
	inter_pnt2.setY(start_pnt2.y() + (end_pnt2.y() - start_pnt2.y()) * Value);

	Viewport.setTopLeft(inter_pnt1);
	Viewport.setBottomRight(inter_pnt2);

	this->fitInView(Viewport,Qt::KeepAspectRatio);

	this->setUpdatesEnabled(true);
}

void TGridWidget::AnimationFinished(void)
{
	d_GraphicsScene->Set_Animation_Mode(false);

	d_AnimationRunning = false;

	d_current_viewport = d_target_viewport;

	this->fitInView(d_current_viewport,Qt::KeepAspectRatio);

	if(d_delayed_remove_widget_active)
	{
		d_delayed_remove_widget_active = false;
		this->Delayed_RemoveWidget();
	}

	emit Animation_Complete();
}

void TGridWidget::wheelEvent(
	QWheelEvent							*Event)
{
	Q_UNUSED(Event);
}

void TGridWidget::mousePressEvent(
	QMouseEvent 						*event)
{
 	QRect								rect;
 	int									row;
 	int									column;
 	int									index;
 	QPointF								viewport_pos;

	QGraphicsView::mousePressEvent(event);


	if(d_CurrentActiveWidgetIndex < 0)
	{
		if(event->button() == Qt::LeftButton)
		{
			viewport_pos = this->mapToScene(event->pos());

			column = viewport_pos.x() / d_WidgetSize.width();
			row = viewport_pos.y() / d_WidgetSize.height();

			if((column < 0) == false && (column < d_MaxColumns) == true)	// check for view
			{
				index = row * d_MaxColumns + column;

				if(static_cast<unsigned int>(index) < d_Widgets.size())
				{
					emit Grid_Item_Selected(d_Widgets[index]->widget());
				}
			}
		}
	}
}

void TGridWidget::Delayed_RemoveWidget(void)
{
	std::vector<QGraphicsProxyWidget*>::iterator proxy_iter;
	std::vector<const QWidget*>::iterator 		 widget_iter;

	for(widget_iter = d_delayed_remove_widgets.begin();widget_iter != d_delayed_remove_widgets.end();widget_iter++)
	{
		for(proxy_iter = d_Widgets.begin();proxy_iter != d_Widgets.end();proxy_iter++)
		{
			if((*widget_iter) == (*proxy_iter)->widget())
			{
				d_GraphicsScene->removeItem((*proxy_iter));
				d_Widgets.erase(proxy_iter);

				break;
			}
		}
	}

	d_delayed_remove_widgets.clear();
	this->Resize_Widgets();
}

void TGridWidget::Create_Background_Image(void)
{
	int									row;
	int									column;
	QSize								size;
	QRectF								target_rect;
	QRect								clear_rect;
	QPixmap								scene_image;
	QBrush								background_brush(this->palette().window());

	row = (d_Widgets.size() - 1) / d_MaxColumns;

	target_rect.setRect(0,0,d_WidgetSize.width() * d_MaxColumns,d_WidgetSize.height() * (row + 1));

	size.setWidth(target_rect.width());
	size.setHeight(target_rect.height());

	scene_image = QPixmap(size);

	if(!scene_image.isNull())
	{
		QPainter 						painter(&scene_image);

		d_GraphicsScene->render(&painter,target_rect,target_rect);

		// Fill in non-used areas
		clear_rect.setRect(0,0,target_rect.width(),target_rect.height());

		row = d_Widgets.size() / d_MaxColumns;
		column = d_Widgets.size() % d_MaxColumns;

		clear_rect.setLeft(column * d_WidgetSize.width());
		clear_rect.setTop(row * d_WidgetSize.height());

		painter.fillRect(clear_rect,background_brush);
	}

	d_GraphicsScene->Set_Background_Image(scene_image);
}



