/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2017  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFont>
#include <QGridLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QToolButton>
#include <QWidget>
#include <QHeaderView>
#include <QSpinBox>
#include <QComboBox>
#include <QClipboard>
#include <QSettings>
#include <QCloseEvent>
#include <QDir>
#include <assert.h>

#include "../../core/tableeditor.h"
#include "../../core/interpolation.h"
#include "../../core/valueedit.h"

#include "interpolationgraphdialog.h"
#include "interpolationeditor.h"

static const double						MINIMUM_INCREMENT(1.0);
static const int						MINIMUM_COUNT(2);


TInterpolationEditor::TInterpolationEditor(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QFont								copyright_font;
	QGridLayout							*edit_layout;
	QGridLayout							*grid_layout;
	QGridLayout							*input_layout;
	QGridLayout							*output_layout;
	QHBoxLayout							*button_hlayout;
	QHBoxLayout							*edit_button_layout;
	QHBoxLayout							*input_button_layout;
	QHBoxLayout							*output_button_hlayout;
	QLabel								*input_count_label;
	QLabel								*edit_count_label;
	QLabel								*output_count_label;
	QLabel								*copyright_label;
	QLabel								*edit_end_label;
	QLabel								*edit_increment_label;
	QLabel								*edit_start_label;
	QLabel								*input_end_label;
	QLabel								*input_increment_label;
	QLabel								*input_start_label;
	QLabel								*output_end_label;
	QLabel								*output_increment_label;
	QLabel								*output_start_label;
	QLabel								*precision_label;
	QSizePolicy							size_policy_preferred(QSizePolicy::Preferred,QSizePolicy::Preferred);
	QSpacerItem							*button_hspacer;
	QSpacerItem							*edit_button_hspacer;
	QSpacerItem							*edit_hspacer;
	QSpacerItem							*input_button_hspacer;
	QSpacerItem							*input_hspacer;
	QSpacerItem							*output_button_hspacer;
	QSpacerItem							*output_hspacer;
	QWidget								*central_widget;
	QDir								exec_path;
	QString								settings_path;

	exec_path = QDir::home();
	
	if(!exec_path.exists(".interpolationeditor"))
	{
		exec_path.mkdir(".interpolationeditor");
	}
	
	exec_path.cd(".interpolationeditor");
	
	settings_path = exec_path.absolutePath();
	
	d_settings = new QSettings(settings_path + "/Settings.ini",QSettings::IniFormat);
	
	size_policy_preferred.setHorizontalStretch(0);
	size_policy_preferred.setVerticalStretch(0);

	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	this->resize(1001,635);

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	grid_layout = new QGridLayout(central_widget);

	
	d_input_group = new QGroupBox(central_widget);
	
	input_layout = new QGridLayout(d_input_group);
	input_layout->setVerticalSpacing(4);
	
	input_start_label = new QLabel(d_input_group);
	input_layout->addWidget(input_start_label,0,0,1,1);
	
	d_input_start_edit = new TDoubleEdit(d_input_group);
	d_input_start_edit->setMinimumSize(100,0);
	d_input_start_edit->setAlignment(Qt::AlignCenter);
	input_layout->addWidget(d_input_start_edit,0,1,1,1);
	
	input_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	input_layout->addItem(input_hspacer,0,2,3,1);
	
	input_increment_label = new QLabel(d_input_group);
	input_layout->addWidget(input_increment_label,1,0,1,1);
	
	d_input_increment_edit = new TDoubleEdit(d_input_group);
	d_input_increment_edit->setMinimumSize(100,0);
	d_input_increment_edit->setAlignment(Qt::AlignCenter);
	input_layout->addWidget(d_input_increment_edit,1,1,1,1);
	
	input_end_label = new QLabel(d_input_group);
	input_layout->addWidget(input_end_label,2,0,1,1);
	
	d_input_end_edit = new TDoubleEdit(d_input_group);
	d_input_end_edit->setMinimumSize(100,0);
	d_input_end_edit->setAlignment(Qt::AlignCenter);
	input_layout->addWidget(d_input_end_edit,2,1,1,1);
	
	input_count_label = new QLabel(d_input_group);
	input_layout->addWidget(input_count_label,3,0,1,1);

	d_input_count_edit = new TIntEdit(d_input_group);
	d_input_count_edit->setMinimumSize(100,0);
	d_input_count_edit->setAlignment(Qt::AlignCenter);
	input_layout->addWidget(d_input_count_edit,3,1,1,1);

	
	d_input_table = new TEditTable(d_input_group);
	input_layout->addWidget(d_input_table,4,0,1,3);
	
	input_button_layout = new QHBoxLayout();
	
	d_input_selection_combo = new QComboBox(d_input_group);
	input_button_layout->addWidget(d_input_selection_combo);

	input_button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	input_button_layout->addItem(input_button_hspacer);
	
	input_layout->addLayout(input_button_layout,5,0,1,3);
	grid_layout->addWidget(d_input_group,0,0,1,1);

	
	d_edit_group = new QGroupBox(central_widget);
	
	edit_layout = new QGridLayout(d_edit_group);
	edit_layout->setVerticalSpacing(4);
	
	edit_start_label = new QLabel(d_edit_group);
	edit_layout->addWidget(edit_start_label,0,0,1,1);
	
	d_edit_start_edit = new TDoubleEdit(d_edit_group);
	d_edit_start_edit->setMinimumSize(100,0);
	d_edit_start_edit->setAlignment(Qt::AlignCenter);
	edit_layout->addWidget(d_edit_start_edit,0,1,1,1);
	
	edit_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	edit_layout->addItem(edit_hspacer,0,2,3,1);

	edit_increment_label = new QLabel(d_edit_group);
	edit_layout->addWidget(edit_increment_label,1,0,1,1);
	
	d_edit_increment_edit = new TDoubleEdit(d_edit_group);
	d_edit_increment_edit->setMinimumSize(100,0);
	d_edit_increment_edit->setAlignment(Qt::AlignCenter);
	edit_layout->addWidget(d_edit_increment_edit,1,1,1,1);
	
	edit_end_label = new QLabel(d_edit_group);
	edit_layout->addWidget(edit_end_label,2,0,1,1);
	
	d_edit_end_edit = new TDoubleEdit(d_edit_group);
	d_edit_end_edit->setMinimumSize(100,0);
	d_edit_end_edit->setAlignment(Qt::AlignCenter);
	edit_layout->addWidget(d_edit_end_edit,2,1,1,1);
	
	edit_count_label = new QLabel(d_input_group);
	edit_layout->addWidget(edit_count_label,3,0,1,1);
	
	d_edit_count_edit = new TIntEdit(d_input_group);
	d_edit_count_edit->setMinimumSize(100,0);
	d_edit_count_edit->setAlignment(Qt::AlignCenter);
	edit_layout->addWidget(d_edit_count_edit,3,1,1,1);
	
	d_edit_table = new TEditTable(d_edit_group);
	edit_layout->addWidget(d_edit_table,4,0,1,3);
	
	edit_button_layout = new QHBoxLayout();
	
	d_edit_selection_combo = new QComboBox(d_edit_group);
	edit_button_layout->addWidget(d_edit_selection_combo);
	
	edit_button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	edit_button_layout->addItem(edit_button_hspacer);
	
	edit_layout->addLayout(edit_button_layout,5,0,1,3);
	grid_layout->addWidget(d_edit_group,0,1,1,1);

	d_output_group = new QGroupBox(central_widget);

	output_layout = new QGridLayout(d_output_group);
	output_layout->setVerticalSpacing(4);

	output_start_label = new QLabel(d_output_group);
	output_layout->addWidget(output_start_label,0,0,1,1);

	d_output_start_edit = new TDoubleEdit(d_output_group);
	d_output_start_edit->setMinimumSize(100,0);
	d_output_start_edit->setAlignment(Qt::AlignCenter);
	d_output_start_edit->setReadOnly(true);
	output_layout->addWidget(d_output_start_edit,0,1,1,1);

	output_increment_label = new QLabel(d_output_group);
	output_layout->addWidget(output_increment_label,1,0,1,1);

	d_output_increment_edit = new TDoubleEdit(d_output_group);
	d_output_increment_edit->setMinimumSize(100,0);
	d_output_increment_edit->setAlignment(Qt::AlignCenter);
	d_output_increment_edit->setReadOnly(true);
	output_layout->addWidget(d_output_increment_edit,1,1,1,1);

	output_end_label = new QLabel(d_output_group);
	output_layout->addWidget(output_end_label,2,0,1,1);

	d_output_end_edit = new TDoubleEdit(d_output_group);
	d_output_end_edit->setMinimumSize(100,0);
	d_output_end_edit->setAlignment(Qt::AlignCenter);
	d_output_end_edit->setReadOnly(true);
	output_layout->addWidget(d_output_end_edit,2,1,1,1);
	
	output_count_label = new QLabel(d_input_group);
	output_layout->addWidget(output_count_label,3,0,1,1);
	
	d_output_count_edit = new TIntEdit(d_input_group);
	d_output_count_edit->setMinimumSize(100,0);
	d_output_count_edit->setAlignment(Qt::AlignCenter);
	d_output_count_edit->setReadOnly(true);
	output_layout->addWidget(d_output_count_edit,3,1,1,1);

	d_output_table = new TEditTable(d_output_group);
	output_layout->addWidget(d_output_table,4,0,1,3);

	output_button_hlayout = new QHBoxLayout();
	
	output_button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	output_button_hlayout->addItem(output_button_hspacer);
	
	d_output_copy_option_combo = new QComboBox(d_output_group);
	output_button_hlayout->addWidget(d_output_copy_option_combo);

	d_output_copy_button = new QToolButton(d_output_group);
	d_output_copy_button->setSizePolicy(size_policy_preferred);
	output_button_hlayout->addWidget(d_output_copy_button);
	output_layout->addLayout(output_button_hlayout,5,0,1,3);

	output_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	output_layout->addItem(output_hspacer,0,2,3,1);
	grid_layout->addWidget(d_output_group,0,2,1,1);

	button_hlayout = new QHBoxLayout();

	copyright_label = new QLabel(central_widget);
	copyright_label->setFont(copyright_font);
	button_hlayout->addWidget(copyright_label);

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);
	
	d_zero_offset_check = new QToolButton(central_widget);
	d_zero_offset_check->setCheckable(true);
	d_zero_offset_check->setSizePolicy(size_policy_preferred);
	button_hlayout->addWidget(d_zero_offset_check);
	
	precision_label = new QLabel(central_widget);
	button_hlayout->addWidget(precision_label);

	d_precision_spin = new QSpinBox(central_widget);
	d_precision_spin->setRange(1,9);
	button_hlayout->addWidget(d_precision_spin);

	d_graph_toggle_button = new QToolButton(central_widget);
	d_graph_toggle_button->setCheckable(true);
	d_graph_toggle_button->setSizePolicy(size_policy_preferred);
	button_hlayout->addWidget(d_graph_toggle_button);
	grid_layout->addLayout(button_hlayout,1,0,1,3);
	
	// extra dialogs
	d_interpolation_graph_dialog = new TInterpolationGraphDialog(this,Qt::Tool);
	
	// defaults
	
	d_input_start_edit->Set_Value(d_settings->value("Input_Start",-10.0).toDouble());
	d_input_increment_edit->Set_Value(d_settings->value("Input_Increment",15.0).toDouble());
	d_input_end_edit->Set_Value(d_settings->value("Input_End",215.0).toDouble());
	
	d_edit_start_edit->Set_Value(d_settings->value("Edit_Start",0.0).toDouble());
	d_edit_increment_edit->Set_Value(d_settings->value("Edit_Increment",50.0).toDouble());
	d_edit_end_edit->Set_Value(d_settings->value("Edit_End",200.0).toDouble());
	
	d_output_start_edit->Set_Value(d_input_start_edit->Value());
	d_output_increment_edit->Set_Value(d_input_increment_edit->Value());
	d_output_end_edit->Set_Value(d_input_end_edit->Value());
	
	d_input_selection_combo->addItem(QStringLiteral("Enter Down"),TEditTable::SELECTION_MODE_DOWN);
	d_input_selection_combo->addItem(QStringLiteral("Enter Up"),TEditTable::SELECTION_MODE_UP);
	
	d_edit_selection_combo->addItem(QStringLiteral("Enter Down"),TEditTable::SELECTION_MODE_DOWN);
	d_edit_selection_combo->addItem(QStringLiteral("Enter Up"),TEditTable::SELECTION_MODE_UP);
	
	d_output_copy_option_combo->addItem(QStringLiteral("Data Only"),TInterpolationEditor::COPY_DATA_ONLY);
	d_output_copy_option_combo->addItem(QStringLiteral("Position and Data"),TInterpolationEditor::COPY_POSITION_AND_DATA);
	
	d_input_selection_combo->setCurrentIndex(d_settings->value("Input_Selection_Mode",0).toInt());
	d_edit_selection_combo->setCurrentIndex(d_settings->value("Edit_Selection_Mode",0).toInt());
	d_output_copy_option_combo->setCurrentIndex(d_settings->value("Output_Copy_Option",0).toInt());
	
	d_input_table->Set_Selection_Mode(TEditTable::SELECTION_MODE_DOWN);
	d_edit_table->Set_Selection_Mode(TEditTable::SELECTION_MODE_DOWN);
	
	d_input_table->Set_Edit_Mode(TEditTable::EDIT_MODE_REPLACE);
	d_edit_table->Set_Edit_Mode(TEditTable::EDIT_MODE_REPLACE);
	d_output_table->Set_Edit_Mode(TEditTable::EDIT_MODE_REPLACE);
	
	d_precision_spin->setValue(d_settings->value("Display_Precision",1).toInt());
	d_zero_offset_check->setChecked(d_settings->value("Zero_Offset_Data",false).toBool());

	
	d_input_table->Enable_Invert_Action(false);
	d_edit_table->Enable_Invert_Action(false);
	d_output_table->Enable_Invert_Action(false);
	d_output_table->Set_Read_Only_Mode(true);

	d_multicell_active = false;

	QWidget::setTabOrder(d_input_start_edit,d_input_increment_edit);
	QWidget::setTabOrder(d_input_increment_edit,d_input_end_edit);
	QWidget::setTabOrder(d_input_end_edit,d_input_count_edit);
	QWidget::setTabOrder(d_input_count_edit,d_input_table);
	QWidget::setTabOrder(d_input_table,d_input_selection_combo);
	QWidget::setTabOrder(d_input_selection_combo,d_edit_start_edit);
	QWidget::setTabOrder(d_edit_start_edit,d_edit_increment_edit);
	QWidget::setTabOrder(d_edit_increment_edit,d_edit_end_edit);
	QWidget::setTabOrder(d_edit_end_edit,d_edit_count_edit);
	QWidget::setTabOrder(d_edit_count_edit,d_edit_table);
	QWidget::setTabOrder(d_edit_table,d_edit_selection_combo);
	QWidget::setTabOrder(d_edit_selection_combo,d_output_start_edit);
	QWidget::setTabOrder(d_output_start_edit,d_output_increment_edit);
	QWidget::setTabOrder(d_output_increment_edit,d_output_end_edit);
	QWidget::setTabOrder(d_output_end_edit,d_output_count_edit);
	QWidget::setTabOrder(d_output_count_edit,d_output_table);
	QWidget::setTabOrder(d_output_table,d_output_copy_option_combo);
	QWidget::setTabOrder(d_output_copy_option_combo,d_output_copy_button);
	QWidget::setTabOrder(d_output_copy_button,d_graph_toggle_button);
	QWidget::setTabOrder(d_output_copy_button,d_zero_offset_check);
	QWidget::setTabOrder(d_zero_offset_check,d_precision_spin);
	
	this->setWindowTitle(QStringLiteral("Interpolation Editor"));
	
	d_output_group->setTitle(QStringLiteral("Output Data"));
	output_start_label->setText(QStringLiteral("Start Position:"));
	output_increment_label->setText(QStringLiteral("Increment:"));
	output_end_label->setText(QStringLiteral("End Position:"));
	output_count_label->setText(QStringLiteral("Count:"));
	d_output_copy_button->setText(QStringLiteral("Copy to Clipboard"));
	d_input_group->setTitle(QStringLiteral("Input Data"));
	input_start_label->setText(QStringLiteral("Start Position:"));
	input_increment_label->setText(QStringLiteral("Increment:"));
	input_end_label->setText(QStringLiteral("End Position:"));
	input_count_label->setText(QStringLiteral("Count:"));
	d_zero_offset_check->setText(QStringLiteral("Zero Offset Data"));
	d_edit_group->setTitle(QStringLiteral("Editor"));
	edit_increment_label->setText(QStringLiteral("Increment:"));
	edit_end_label->setText(QStringLiteral("End Position:"));
	edit_count_label->setText(QStringLiteral("Count:"));
	edit_start_label->setText(QStringLiteral("Start Position:"));
	copyright_label->setText(QStringLiteral("Copyright (C) 2017 Select Calibration Incorporated.  www.selectcalibration.ca"));
	d_graph_toggle_button->setText(QStringLiteral("Show Graph"));
	precision_label->setText(QStringLiteral("Display Precision:"));

	this->Initialize_Input_Output_Editor();
	this->Initialize_Edit_Editor();
	this->Update_Graph();
	
	this->Input_Selection_Mode_Changed();
	this->Edit_Selection_Mode_Changed();
	this->Display_Precision_Changed(d_precision_spin->value());

	connect(d_precision_spin,SIGNAL(valueChanged(int)),this,SLOT(Display_Precision_Changed(int)));
	connect(d_output_copy_button,SIGNAL(clicked(void)),this,SLOT(Copy_Output(void)));
	connect(d_graph_toggle_button,SIGNAL(toggled(bool)),this,SLOT(Show_Hide_Graph(bool)));
	connect(d_zero_offset_check,SIGNAL(toggled(bool)),this,SLOT(Zero_Offset_Toggled(bool)));
	connect(d_interpolation_graph_dialog,SIGNAL(Dialog_Close(void)),this,SLOT(Graph_Closed(void)));
	
	connect(d_input_selection_combo,SIGNAL(currentIndexChanged(int)),this,SLOT(Input_Selection_Mode_Changed(void)));
	connect(d_input_start_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Input_Range_Changed(void)));
	connect(d_input_increment_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Input_Increment_Count_Changed(void)));
	connect(d_input_end_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Input_Range_Changed(void)));
	connect(d_input_count_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Input_Increment_Count_Changed(void)));
	
	connect(d_edit_selection_combo,SIGNAL(currentIndexChanged(int)),this,SLOT(Edit_Selection_Mode_Changed(void)));
	connect(d_edit_start_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Edit_Range_Changed(void)));
	connect(d_edit_increment_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Edit_Increment_Count_Changed(void)));
	connect(d_edit_end_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Edit_Range_Changed(void)));
	connect(d_edit_count_edit,SIGNAL(Edit_Finished(void)),this,SLOT(Edit_Increment_Count_Changed(void)));

	connect(d_input_table,SIGNAL(Cell_Changed(const int,const int,const double&)),this,SLOT(Input_Changed(const int,const int,const double&)));
	connect(d_input_table,SIGNAL(Start_Multiple_Cell_Change(const QString&)),this,SLOT(Start_Multiple_Cell_Change(void)));
	connect(d_input_table,SIGNAL(End_Multiple_Cell_Change(void)),this,SLOT(End_Multiple_Cell_Change(void)));

	connect(d_edit_table,SIGNAL(Cell_Changed(const int,const int,const double&)),this,SLOT(Edit_Changed(const int,const int,const double&)));
	connect(d_edit_table,SIGNAL(Start_Multiple_Cell_Change(const QString&)),this,SLOT(Start_Multiple_Cell_Change(void)));
	connect(d_edit_table,SIGNAL(End_Multiple_Cell_Change(void)),this,SLOT(End_Multiple_Cell_Change(void)));
}

TInterpolationEditor::~TInterpolationEditor(void)
{
}

void TInterpolationEditor::Display_Precision_Changed(
	int									precision)
{
	d_input_table->Set_Display_Precision(precision);
	d_edit_table->Set_Display_Precision(precision);
	d_output_table->Set_Display_Precision(precision);
	
	d_input_start_edit->Set_Display_Precision(precision);
	d_input_increment_edit->Set_Display_Precision(precision);
	d_input_end_edit->Set_Display_Precision(precision);
	
	d_edit_start_edit->Set_Display_Precision(precision);
	d_edit_increment_edit->Set_Display_Precision(precision);
	d_edit_end_edit->Set_Display_Precision(precision);
	
	d_output_start_edit->Set_Display_Precision(precision);
	d_output_increment_edit->Set_Display_Precision(precision);
	d_output_end_edit->Set_Display_Precision(precision);
}
/*
void TInterpolationEditor::Discard_Edit(void)
{
	d_input_group->setEnabled(true);
}
*/
void TInterpolationEditor::Copy_Output(void)
{
	int									precision;
	QString								text;
	std::vector<TVector2>::iterator		iter;
	TInterpolationEditor::TCopyOption	copy_option;
	
	copy_option = static_cast<TInterpolationEditor::TCopyOption>(d_output_copy_option_combo->currentData().toInt());
	precision = d_precision_spin->value();
	
	if(copy_option == TInterpolationEditor::COPY_DATA_ONLY)
	{
		for(iter = d_output_data.begin();iter != d_output_data.end();++iter)
		{
			text += QString("%1\n").arg((*iter).y,0,'f',precision);
		}
	}
	else	// assume position and values
	{
		for(iter = d_output_data.begin();iter != d_output_data.end();++iter)
		{
			text += QString("%1\t%2\n").arg((*iter).x,0,'f',precision).arg((*iter).y,0,'f',precision);
		}
	}
	
	QApplication::clipboard()->setText(text,QClipboard::Clipboard);
}

void TInterpolationEditor::Show_Hide_Graph(
	bool								state)
{
	d_interpolation_graph_dialog->setVisible(state);
	
	if(state)
	{
		d_graph_toggle_button->setText(QStringLiteral("Hide Graph"));
	}
	else
	{
		d_graph_toggle_button->setText(QStringLiteral("Show Graph"));
	}
}

void TInterpolationEditor::Graph_Closed(void)
{
	d_graph_toggle_button->setChecked(false);
}

void TInterpolationEditor::Zero_Offset_Toggled(
	bool								state)
{
	Q_UNUSED(state);
	
	this->Update_Output_Editor();
	this->Update_Graph();
}

void TInterpolationEditor::Input_Selection_Mode_Changed(void)
{
	TEditTable::TSelectionMode			mode;
	
	mode = static_cast<TEditTable::TSelectionMode>(d_input_selection_combo->currentData().toInt());
	d_input_table->Set_Selection_Mode(mode);
}

void TInterpolationEditor::Input_Range_Changed(void)
{
	// output table start and end must match input start and end
	d_output_start_edit->Set_Value(d_input_start_edit->Value());
	d_output_end_edit->Set_Value(d_input_end_edit->Value());
	
	this->Initialize_Input_Output_Editor();
	this->Update_Graph();
}

void TInterpolationEditor::Input_Increment_Count_Changed(void)
{
	double								start;
	double								increment;
	double								end;
	int									count;
	
	// calculate end based on count, start, and increment
	
	start = d_input_start_edit->Value();
	increment = d_input_increment_edit->Value();
	count = d_input_count_edit->Value();
	
	if(count < MINIMUM_COUNT)
	{
		count = MINIMUM_COUNT;
		d_input_count_edit->Set_Value(count);
		d_output_count_edit->Set_Value(count);
	}
	
	d_output_increment_edit->Set_Value(increment);
	
	end = start + increment * static_cast<double>(count - 1);
	
	d_input_end_edit->Set_Value(end);
	d_output_end_edit->Set_Value(end);
	
	this->Initialize_Input_Output_Editor();
	
}

void TInterpolationEditor::Input_Changed(
	const int							row,
	const int							column,
	const double						&value)
{
	assert(column == 0);
	assert(static_cast<unsigned int>(row) < d_input_data.size());
	
	d_input_data[row].y = value;
	
	if(!d_multicell_active)
	{
		this->Update_Output_Editor();
		this->Update_Graph();
	}
}

void TInterpolationEditor::Edit_Selection_Mode_Changed(void)
{
	TEditTable::TSelectionMode			mode;
	
	mode = static_cast<TEditTable::TSelectionMode>(d_edit_selection_combo->currentData().toInt());
	d_edit_table->Set_Selection_Mode(mode);
}

void TInterpolationEditor::Edit_Range_Changed(void)
{
	this->Initialize_Edit_Editor();
}

void TInterpolationEditor::Edit_Increment_Count_Changed(void)
{
	double								start;
	double								increment;
	double								end;
	int									count;
	
	// calculate end based on count, start, and increment
	
	start = d_edit_start_edit->Value();
	increment = d_edit_increment_edit->Value();
	count = d_edit_count_edit->Value();
	
	if(count < MINIMUM_COUNT)
	{
		count = MINIMUM_COUNT;
		d_edit_count_edit->Set_Value(count);
	}
	
	end = start + increment * static_cast<double>(count - 1);
	d_edit_end_edit->Set_Value(end);
	
	this->Initialize_Edit_Editor();
}

void TInterpolationEditor::Edit_Changed(
	const int							row,
	const int							column,
	const double						&value)
{
	assert(column == 0);
	assert(static_cast<unsigned int>(row) < d_edit_data.size());
	
//	d_input_group->setEnabled(false);
	
	d_edit_data[row].y = value;
	
	if(!d_multicell_active)
	{
		this->Update_Output_Editor();
		this->Update_Graph();
	}
}

void TInterpolationEditor::Start_Multiple_Cell_Change(void)
{
	d_multicell_active = true;
}

void TInterpolationEditor::End_Multiple_Cell_Change(void)
{
	d_multicell_active = false;
	
	this->Update_Output_Editor();
	this->Update_Graph();
}

void TInterpolationEditor::closeEvent(
	QCloseEvent							*event)
{
	d_settings->setValue("Input_Start", d_input_start_edit->Value());
	d_settings->setValue("Input_Increment", d_input_increment_edit->Value());
	d_settings->setValue("Input_End", d_input_end_edit->Value());
	
	d_settings->setValue("Edit_Start", d_edit_start_edit->Value());
	d_settings->setValue("Edit_Increment", d_edit_increment_edit->Value());
	d_settings->setValue("Edit_End", d_edit_end_edit->Value());
	
	d_settings->setValue("Input_Selection_Mode", d_input_selection_combo->currentIndex());
	d_settings->setValue("Edit_Selection_Mode", d_edit_selection_combo->currentIndex());
	d_settings->setValue("Output_Copy_Option", d_output_copy_option_combo->currentIndex());
	
	d_settings->setValue("Display_Precision", d_precision_spin->value());
	d_settings->setValue("Zero_Offset_Data", d_zero_offset_check->isChecked());

	event->accept();
}

void TInterpolationEditor::Initialize_Input_Output_Editor(void)
{
	double								start;
	double								increment;
	double								end;
	double								position;
	int									count;
	int									cntr;
	std::vector<TVector2>::iterator		iter;
	
	start = d_input_start_edit->Value();
	increment = d_input_increment_edit->Value();
	end = d_input_end_edit->Value();
	
	// verify data inputs, change increment if wrong sign
	if(end > start)
	{
		if(increment < MINIMUM_INCREMENT)	// verify both sign and value
		{
			increment = MINIMUM_INCREMENT;

			d_input_increment_edit->Set_Value(increment);
			d_output_increment_edit->Set_Value(increment);
		}
		
		count = 1 + static_cast<int>((end - start) / increment + 0.5);
		
		if(!(count < MINIMUM_COUNT))
		{
			d_input_table->Set_Cell_Count(count,1);
			d_output_table->Set_Cell_Count(count,1);
		}
		else
		{
			d_input_table->Set_Cell_Count(0,0);
			d_output_table->Set_Cell_Count(0,0);
		}
	}
	else if(end < start)
	{
		if(increment > (-MINIMUM_INCREMENT))	// verify both sign and value
		{
			increment = (-MINIMUM_INCREMENT);
			
			d_input_increment_edit->Set_Value(increment);
			d_output_increment_edit->Set_Value(increment);
		}
		
		count = 1 + static_cast<int>((end - start) / increment + 0.5);
		
		if(!(count < MINIMUM_COUNT))
		{
			d_input_table->Set_Cell_Count(count,1);
			d_output_table->Set_Cell_Count(count,1);
		}
		else
		{
			d_input_table->Set_Cell_Count(0,0);
			d_output_table->Set_Cell_Count(0,0);
		}
		
	}
	else	// start and end are the same
	{
		d_input_table->Set_Cell_Count(0,0);
		d_output_table->Set_Cell_Count(0,0);
		return;
	}
	
	d_input_data.resize(count);
	
	d_input_count_edit->Set_Value(count);
	d_output_count_edit->Set_Value(count);
	
	position = start;
	for(cntr = 0,iter = d_input_data.begin();cntr < count;++cntr,++iter)
	{
		(*iter).x = position;
		
		d_input_table->verticalHeaderItem(cntr)->setText(QString("%1").arg(position,0,'f',3));
		d_output_table->verticalHeaderItem(cntr)->setText(QString("%1").arg(position,0,'f',3));

		d_input_table->Set_Cell_Value(cntr,0,(*iter).y);
		d_output_table->Set_Cell_Value(cntr,0,(*iter).y);
		
		position += increment;
	}
	
	d_output_data = d_input_data;
	
	d_input_table->horizontalHeaderItem(0)->setText(QStringLiteral("Value"));
	d_output_table->horizontalHeaderItem(0)->setText(QStringLiteral("Value"));
}

void TInterpolationEditor::Initialize_Edit_Editor(void)
{
	double								start;
	double								increment;
	double								end;
	double								position;
	double								value;
	int									count;
	int									cntr;
	std::vector<TVector2>::iterator		iter;
	double								minimum_increment;
	
	start = d_edit_start_edit->Value();
	increment = d_edit_increment_edit->Value();
	end = d_edit_end_edit->Value();
	
	minimum_increment = MINIMUM_INCREMENT;
	value = fabs(d_input_increment_edit->Value());
	
	if(value > minimum_increment)
	{
		minimum_increment = value;
	}
	
	// verify data inputs, change increment if wrong sign
	if(end > start)
	{
		if(increment < minimum_increment)	// verify both sign and value
		{
			increment = minimum_increment;
			
			d_edit_increment_edit->Set_Value(increment);
		}
		
		count = 1 + static_cast<int>((end - start) / increment + 0.5);
		
		if(!(count < MINIMUM_COUNT))
		{
			d_edit_table->Set_Cell_Count(count,1);
		}
		else
		{
			d_edit_table->Set_Cell_Count(0,0);
		}
	}
	else if(end < start)
	{
		if(increment > (-minimum_increment))	// verify both sign and value
		{
			increment = (-minimum_increment);
			
			d_edit_increment_edit->Set_Value(increment);
		}
		
		count = 1 + static_cast<int>((end - start) / increment + 0.5);
		
		if(!(count < MINIMUM_COUNT))
		{
			d_edit_table->Set_Cell_Count(count,1);
		}
		else
		{
			d_edit_table->Set_Cell_Count(0,0);
		}
		
	}
	else	// start and end are the same
	{
		d_edit_table->Set_Cell_Count(0,0);
		return;
	}

	d_edit_data.resize(count);

	d_edit_count_edit->Set_Value(count);

	position = start;
	for(cntr = 0,iter = d_edit_data.begin();cntr < count;++cntr,++iter)
	{
		(*iter).x = position;
		(*iter).y = 0.0;
		
		d_edit_table->verticalHeaderItem(cntr)->setText(QString("%1").arg((*iter).x,0,'f',3));
		d_edit_table->Set_Cell_Value(cntr,0,(*iter).y);
		
		position += increment;
	}
	
	d_edit_table->horizontalHeaderItem(0)->setText(QStringLiteral("Value"));
}

void TInterpolationEditor::Update_Edit_Editor(void)
{
	int									cntr;
	std::vector<TVector2>::iterator		iter;
		
	assert(static_cast<unsigned int>(d_edit_table->Row_Count()) == d_edit_data.size());

	for(iter = d_edit_data.begin(),cntr = 0;iter != d_edit_data.end();++iter,++cntr)
	{
		Extrapolation::Get_Value(d_input_data,(*iter).x,&(*iter).y);
		d_edit_table->Set_Cell_Value(cntr,0,(*iter).y);
	}
}

void TInterpolationEditor::Update_Output_Editor(void)
{
	int									cntr;
	double								value;
	double								zero_offset;
	std::vector<TVector2>::iterator		in_iter,out_iter;
	
	assert(static_cast<unsigned int>(d_output_table->Row_Count()) == d_input_data.size());
	assert(d_output_data.size() == d_input_data.size());
	
	for(in_iter = d_input_data.begin(),out_iter = d_output_data.begin(),cntr = 0;
		in_iter != d_input_data.end();++in_iter,++out_iter,++cntr)
	{
		Extrapolation::Get_Value(d_edit_data,(*in_iter).x,&value);
		
		(*out_iter).y = (*in_iter).y + value;
		d_output_table->Set_Cell_Value(cntr,0,(*out_iter).y);
	}
	
	if(d_zero_offset_check->isChecked())
	{
		Extrapolation::Get_Value(d_output_data,0.0,&zero_offset);
		
		for(out_iter = d_output_data.begin(),cntr = 0;out_iter != d_output_data.end();++out_iter,++cntr)
		{
			(*out_iter).y -= zero_offset;
			d_output_table->Set_Cell_Value(cntr,0,(*out_iter).y);
		}
	}
}

void TInterpolationEditor::Update_Graph(void)
{
	d_interpolation_graph_dialog->Update(d_input_data,d_output_data);
}

